@@ -1,5 +1,5 @@
 import math
-import pywaves
+import pywaves as pywaves
 import axolotl_curve25519 as curve
 import os
 import pywaves.crypto as crypto
@@ -12,6 +12,8 @@ import logging
 
 from .protobuf import transaction_pb2
 from .protobuf.waves import amount_pb2
+from .protobuf.waves import recipient_pb2
+
 from google.protobuf.json_format import MessageToJson
 
 wordList = ['abandon', 'ability', 'able', 'about', 'above', 'absent', 'absorb', 'abstract', 'absurd', 'abuse', 'access',
@@ -261,7 +263,6 @@ class Address(object):
                     '/addresses/balance/%s%s' % (self.address, '' if confirmations == 0 else '/%d' % confirmations))[
                     'balance']
         except Exception as e:
-            print(e)
             return -1
 
     def transactions(self, limit=100, after=''):
@@ -279,6 +280,9 @@ class Address(object):
                 a[i] = a[i][8:]
         return a
 
+    def script(self):
+        return self.pywaves.wrapper('/addresses/scriptInfo/%s' % self.address)
+
     def _generate(self, publicKey='', privateKey='', seed='', nonce=0):
         self.seed = seed
         self.nonce = nonce
@@ -428,6 +432,13 @@ class Address(object):
         else:
             if timestamp == 0:
                 timestamp = int(time.time() * 1000)
+
+            '''transferTransaction = transaction_pb2.TransferTransactionData()
+            transferTransaction.recipient.public_key_hash = crypto.str2bytes(recipient.address)
+            transferTransaction.amount.asset_id = b''
+            transferTransaction.amount.amount = amount
+            transferTransaction.attachment.string_value = base58.b58encode(crypto.str2bytes(attachment))'''
+
             sData = b'\4' + \
                     b'\2' + \
                     base58.b58decode(self.publicKey) + \
@@ -439,6 +450,16 @@ class Address(object):
                     struct.pack(">H", len(attachment)) + \
                     crypto.str2bytes(attachment)
             signature = crypto.sign(self.privateKey, sData)
+            '''txFeeEntry = amount_pb2.Amount()
+            txFeeEntry.amount = txFee
+            transaction = transaction_pb2.Transaction()
+            transaction.chain_id = ord(self.pywaves.CHAIN_ID)
+            transaction.sender_public_key = base58.b58decode(self.publicKey)
+            transaction.fee.CopyFrom(txFeeEntry)
+            transaction.timestamp = timestamp
+            transaction.version = 3
+            transaction.transfer.CopyFrom(transferTransaction)
+            signature = crypto.sign(self.privateKey, transaction.SerializeToString())'''
             data = json.dumps({
                 "type": 4,
                 "version": 2,
@@ -448,7 +469,6 @@ class Address(object):
                 "fee": txFee,
                 "timestamp": timestamp,
                 "attachment": base58.b58encode(crypto.str2bytes(attachment)),
-                "signature": signature,
                 "proofs": [signature]
             })
 
@@ -456,6 +476,7 @@ class Address(object):
 
     def massTransferWaves(self, transfers, attachment='', timestamp=0, baseFee=pywaves.DEFAULT_BASE_FEE):
         txFee = baseFee + (math.ceil((len(transfers) + 1) / 2 - 0.5)) * baseFee
+        txFee += self.script()['extraFee']
         totalAmount = 0
 
         for i in range(0, len(transfers)):
@@ -511,8 +532,8 @@ class Address(object):
 
     def sendAsset(self, recipient, asset, amount, attachment='', feeAsset='', txFee=pywaves.DEFAULT_TX_FEE,
                   timestamp=0):
-        if not self.privateKey:
-            msg = 'Asset not issued'
+        if len(self.privateKey) < 10:
+            msg = 'Private key required'
             logging.error(msg)
             self.pywaves.throw_error(msg)
         elif not self.pywaves.OFFLINE and asset and not asset.status():
@@ -531,12 +552,8 @@ class Address(object):
             msg = 'Insufficient Waves balance'
             logging.error(msg)
             self.pywaves.throw_error(msg)
-        elif not self.pywaves.OFFLINE and not feeAsset and self.balance() < txFee:
-            msg = 'Insufficient Waves balance'
-            logging.error(msg)
-            self.pywaves.throw_error(msg)
         elif not self.pywaves.OFFLINE and feeAsset and self.balance(feeAsset.assetId) < txFee:
-            msg = 'Insufficient asset balance'
+            msg = 'Insufficient asset balance for fee'
             logging.error(msg)
             self.pywaves.throw_error(msg)
         else:
@@ -581,16 +598,28 @@ class Address(object):
         if (asset.isSmart()):
             txFee += smartFee
 
+        txFee += self.script()['extraFee']
+
         totalAmount = 0
 
+        for transfer in transfers:
+            totalAmount += transfer['amount']
         if not self.privateKey:
-            logging.error('Private key required')
+            msg = 'Private key required'
+            logging.error(msg)
+            self.pywaves.throw_error(msg)
         elif len(transfers) > 100:
-            logging.error('Too many recipients')
+            msg = 'Too many recipients'
+            logging.error(msg)
+            self.pywaves.throw_error(msg)
         elif not self.pywaves.OFFLINE and self.balance() < txFee:
-            logging.error('Insufficient Waves balance')
-        elif not self.pywaves.OFFLINE and self.balance(assetId=asset.assetId) < totalAmount:
-            logging.error('Insufficient %s balance' % asset.name)
+            msg = 'Insufficient Waves balance'
+            logging.error(msg)
+            self.pywaves.throw_error(msg)
+        if not self.pywaves.OFFLINE and self.balance(assetId=asset.assetId) < totalAmount:
+            msg = 'Insufficient %s balance' % asset.name
+            logging.error(msg)
+            self.pywaves.throw_error(msg)
         else:
             if timestamp == 0:
                 timestamp = int(time.time() * 1000)
@@ -628,66 +657,6 @@ class Address(object):
 
             return self.pywaves.wrapper('/transactions/broadcast', data)
 
-    '''def dataTransaction(self, data, timestamp=0, baseFee=pywaves.DEFAULT_BASE_FEE, minimalFee=500000):
-        if not self.privateKey:
-            logging.error('Private key required')
-        else:
-            if timestamp == 0:
-                timestamp = int(time.time() * 1000)
-            dataObject = {
-                "type": 12,
-                "version": 1,
-                "senderPublicKey": self.publicKey,
-                "data": data,
-                "fee": 0,
-                "timestamp": timestamp,
-                "proofs": ['']
-            }
-            dataBinary = b''
-            for i in range(0, len(data)):
-                d = data[i]
-                keyBytes = crypto.str2bytes(d['key'])
-                dataBinary += struct.pack(">H", len(keyBytes))
-                dataBinary += keyBytes
-                if d['type'] == 'binary':
-                    dataBinary += b'\2'
-                    valueAsBytes = d['value']
-                    dataBinary += struct.pack(">H", len(valueAsBytes))
-                    dataBinary += crypto.str2bytes(valueAsBytes)
-                elif d['type'] == 'boolean':
-                    if d['value']:
-                        dataBinary += b'\1\1'
-                    else:
-                        dataBinary += b'\1\0'
-                elif d['type'] == 'integer':
-                    dataBinary += b'\0'
-                    dataBinary += struct.pack(">Q", d['value'])
-                elif d['type'] == 'string':
-                    dataBinary += b'\3'
-                    dataBinary += struct.pack(">H", len(d['value']))
-                    dataBinary += crypto.str2bytes(d['value'])
-            # check: https://stackoverflow.com/questions/2356501/how-do-you-round-up-a-number-in-python
-            txFee = (int(((len(crypto.str2bytes(json.dumps(data))) + 2 + 64)) / 1000.0) + 1) * baseFee
-            txFee = max(txFee, minimalFee)
-            dataObject['fee'] = txFee
-            sData = b'\x0c' + \
-                    b'\1' + \
-                    base58.b58decode(self.publicKey) + \
-                    struct.pack(">H", len(data)) + \
-                    dataBinary + \
-                    struct.pack(">Q", timestamp) + \
-                    struct.pack(">Q", txFee)
-
-            dataObject['proofs'] = [crypto.sign(self.privateKey, sData)]
-
-            for entry in dataObject['data']:
-                if entry['type'] == 'binary':
-                    base64Encoded = base64.b64encode(crypto.str2bytes(entry['value']))
-                    entry['value'] = 'base64:' + crypto.bytes2str(base64Encoded)
-            dataObjectJSON = json.dumps(dataObject)
-
-            return self.pywaves.wrapper('/transactions/broadcast', dataObjectJSON)'''
-
     def dataTransaction(self, data, timestamp=0, baseFee=pywaves.DEFAULT_BASE_FEE, minimalFee=500000):
         dataTransaction = transaction_pb2.DataTransactionData()
 
@@ -844,9 +813,6 @@ class Address(object):
             elif req['status'] == 'OrderAccepted':
                 id = req['message']['id']
                 logging.info('Order Accepted - ID: %s' % id)
-        elif not self.pywaves.OFFLINE:
-            logging.error(req)
-            self.pywaves.throw_error(req)
         else:
             return req
         return id
@@ -857,8 +823,7 @@ class Address(object):
                 msg = "Order already filled"
                 logging.error(msg)
                 self.pywaves.throw_error(msg)
-
-            elif not order.status():
+            elif order.status() == 'NotFound':
                 msg = "Order not found"
                 logging.error(msg)
                 self.pywaves.throw_error(msg)
@@ -984,40 +949,6 @@ class Address(object):
             req = self.pywaves.wrapper('/transactions/broadcast', data)
             return req
 
-    '''def lease(self, recipient, amount, txFee=pywaves.DEFAULT_LEASE_FEE, timestamp=0):
-        if not self.privateKey:
-            msg = 'Private key required'
-            logging.error(msg)
-            self.pywaves.throw_error(msg)
-        elif amount <= 0:
-            msg = 'Amount must be > 0'
-            logging.error(msg)
-            self.pywaves.throw_error(msg)
-        elif not self.pywaves.OFFLINE and self.balance() < amount + txFee:
-            msg = 'Insufficient Waves balance'
-            logging.error(msg)
-            self.pywaves.throw_error(msg)
-        else:
-            if timestamp == 0:
-                timestamp = int(time.time() * 1000)
-            sData = b'\x08' + \
-                    base58.b58decode(self.publicKey) + \
-                    base58.b58decode(recipient.address) + \
-                    struct.pack(">Q", amount) + \
-                    struct.pack(">Q", txFee) + \
-                    struct.pack(">Q", timestamp)
-            signature = crypto.sign(self.privateKey, sData)
-            data = json.dumps({
-                "senderPublicKey": self.publicKey,
-                "recipient": recipient.address,
-                "amount": amount,
-                "fee": txFee,
-                "timestamp": timestamp,
-                "signature": signature
-            })
-            req = self.pywaves.wrapper('/leasing/broadcast/lease', data)
-            return req'''
-
     def leaseCancel(self, leaseId, txFee=pywaves.DEFAULT_LEASE_FEE, timestamp=0):
         if not self.privateKey:
             msg = 'Private key required'
@@ -1043,7 +974,8 @@ class Address(object):
                 "timestamp": timestamp,
                 "signature": signature
             })
-            req = self.pywaves.wrapper('/leasing/broadcast/cancel', data)
+            #req = self.pywaves.wrapper('/leasing/broadcast/cancel', data)
+            req = self.pywaves.wrapper('/transactions/broadcast', data)
             if self.pywaves.OFFLINE:
                 return req
             elif 'leaseId' in req:
@@ -1085,22 +1017,22 @@ class Address(object):
                 self.pywaves.DEFAULT_CURRENCY if assetPair.asset2.assetId == '' else assetPair.asset2.assetId), data,
                                      host=self.pywaves.MATCHER)
 
-    def deleteOrderHistory(self, assetPair):
-        orders = self.getOrderHistory(assetPair)
-        for order in orders:
-            orderId = order['id']
-            sData = base58.b58decode(self.publicKey) + \
-                    base58.b58decode(orderId)
-            signature = crypto.sign(self.privateKey, sData)
-            data = json.dumps({
-                "sender": self.publicKey,
-                "orderId": orderId,
-                "signature": signature
-            })
-            self.pywaves.wrapper('/matcher/orderbook/%s/%s/delete' % (
-            self.pywaves.DEFAULT_CURRENCY if assetPair.asset1.assetId == '' else assetPair.asset1.assetId,
-            self.pywaves.DEFAULT_CURRENCY if assetPair.asset2.assetId == '' else assetPair.asset2.assetId), data,
-                                 host=self.pywaves.MATCHER)
+   # def deleteOrderHistory(self, assetPair):
+   #     orders = self.getOrderHistory(assetPair)
+   #     for order in orders:
+   #         orderId = order['id']
+   #         sData = base58.b58decode(self.publicKey) + \
+   #                 base58.b58decode(orderId)
+   #         signature = crypto.sign(self.privateKey, sData)
+   #         data = json.dumps({
+   #             "sender": self.publicKey,
+   #             "orderId": orderId,
+   #             "signature": signature
+   #         })
+   #         tx = self.pywaves.wrapper('/matcher/orderbook/%s/%s/delete' % (
+   #         self.pywaves.DEFAULT_CURRENCY if assetPair.asset1.assetId == '' else assetPair.asset1.assetId,
+   #         self.pywaves.DEFAULT_CURRENCY if assetPair.asset2.assetId == '' else assetPair.asset2.assetId), data,
+   #                              host=self.pywaves.MATCHER)
 
     def createAlias(self, alias, txFee=pywaves.DEFAULT_ALIAS_FEE, timestamp=0):
         aliasWithNetwork = b'\x02' + crypto.str2bytes(str(self.pywaves.CHAIN_ID)) + struct.pack(">H", len(
@@ -1112,12 +1044,6 @@ class Address(object):
         else:
             if timestamp == 0:
                 timestamp = int(time.time() * 1000)
-            '''sData = b'\x0a' + \
-                    base58.b58decode(self.publicKey) + \
-                    struct.pack(">H", len(aliasWithNetwork)) + \
-                    crypto.str2bytes(str(aliasWithNetwork)) + \
-                    struct.pack(">Q", txFee) + \
-                    struct.pack(">Q", timestamp)'''
             sData = b'\x0a' + \
                     base58.b58decode(self.publicKey) + \
                     struct.pack(">H", len(aliasWithNetwork)) + \
@@ -1138,7 +1064,9 @@ class Address(object):
 
     def sponsorAsset(self, assetId, minimalFeeInAssets, txFee=pywaves.DEFAULT_SPONSOR_FEE, timestamp=0):
         if not self.privateKey:
-            logging.error('Private key required')
+            msg = 'Private key required'
+            logging.error(msg)
+            self.pywaves.throw_error(msg)
         else:
             if timestamp == 0:
                 timestamp = int(time.time() * 1000)
@@ -1166,19 +1094,24 @@ class Address(object):
 
             return self.pywaves.wrapper('/transactions/broadcast', data)
 
-    def setScript(self, scriptSource, txFee=pywaves.DEFAULT_SCRIPT_FEE, timestamp=0):
-        script = self.pywaves.wrapper('/utils/script/compile', scriptSource)['script'][7:]
+    def setCompiledScript(self, script, txFee=pywaves.DEFAULT_SCRIPT_FEE, timestamp=0, publicKey=None):
+        #script = self.pywaves.wrapper('/utils/script/compileCode', scriptSource)['script'][7:]
         if not self.privateKey:
-            logging.error('Private key required')
+            msg = 'Private key required'
+            logging.error(msg)
+            self.pywaves.throw_error(msg)
         else:
             compiledScript = base64.b64decode(script)
             scriptLength = len(compiledScript)
             if timestamp == 0:
                 timestamp = int(time.time() * 1000)
+
+            if not publicKey:
+                publicKey = self.publicKey
             sData = b'\x0d' + \
                     b'\1' + \
                     crypto.str2bytes(str(self.pywaves.CHAIN_ID)) + \
-                    base58.b58decode(self.publicKey) + \
+                    base58.b58decode(publicKey) + \
                     b'\1' + \
                     struct.pack(">H", scriptLength) + \
                     compiledScript + \
@@ -1189,7 +1122,7 @@ class Address(object):
             data = json.dumps({
                 "type": 13,
                 "version": 1,
-                "senderPublicKey": self.publicKey,
+                "senderPublicKey": publicKey,
                 "fee": txFee,
                 "timestamp": timestamp,
                 "script": 'base64:' + script,
@@ -1197,13 +1130,54 @@ class Address(object):
                     signature
                 ]
             })
-
             return self.pywaves.wrapper('/transactions/broadcast', data)
 
+    def setScript(self, scriptSource, txFee=pywaves.DEFAULT_SCRIPT_FEE, timestamp=0, publicKey=None):
+        script = self.pywaves.wrapper('/utils/script/compileCode', scriptSource)['script'][7:]
+
+        return self.setCompiledScript(script, txFee, timestamp, publicKey)
+        '''if not self.privateKey:
+            msg = 'Private key required'
+            logging.error(msg)
+            self.pywaves.throw_error(msg)
+        else:
+            compiledScript = base64.b64decode(script)
+            scriptLength = len(compiledScript)
+            if timestamp == 0:
+                timestamp = int(time.time() * 1000)
+
+            if not publicKey:
+                publicKey = self.publicKey
+            sData = b'\x0d' + \
+                    b'\1' + \
+                    crypto.str2bytes(str(self.pywaves.CHAIN_ID)) + \
+                    base58.b58decode(publicKey) + \
+                    b'\1' + \
+                    struct.pack(">H", scriptLength) + \
+                    compiledScript + \
+                    struct.pack(">Q", txFee) + \
+                    struct.pack(">Q", timestamp)
+            signature = crypto.sign(self.privateKey, sData)
+
+            data = json.dumps({
+                "type": 13,
+                "version": 1,
+                "senderPublicKey": publicKey,
+                "fee": txFee,
+                "timestamp": timestamp,
+                "script": 'base64:' + script,
+                "proofs": [
+                    signature
+                ]
+            })
+            return self.pywaves.wrapper('/transactions/broadcast', data)'''
+
     def setAssetScript(self, asset, scriptSource, txFee=pywaves.DEFAULT_ASSET_SCRIPT_FEE, timestamp=0):
         script = self.pywaves.wrapper('/utils/script/compile', scriptSource)['script'][7:]
         if not self.privateKey:
-            logging.error('Private key required')
+            msg = 'Private key required'
+            logging.error(msg)
+            self.pywaves.throw_error(msg)
         else:
             compiledScript = base64.b64decode(script)
             scriptLength = len(compiledScript)
@@ -1238,7 +1212,11 @@ class Address(object):
 
     def issueSmartAsset(self, name, description, quantity, scriptSource, decimals=0, reissuable=False,
                         txFee=pywaves.DEFAULT_ASSET_FEE):
-        script = self.pywaves.wrapper('/utils/script/compile', scriptSource)['script'][7:]
+        if self.pywaves.OFFLINE:
+            msg = 'PyWaves currently offline'
+            logging.error(msg)
+            self.pywaves.throw_error(msg)
+        script = self.pywaves.wrapper('/utils/script/compileCode', scriptSource)['script'][7:]
         if not self.privateKey:
             msg = 'Private key required'
             logging.error(msg)
@@ -1282,11 +1260,7 @@ class Address(object):
                 "proofs": [signature],
                 "script": 'base64:' + script
             })
-            req = self.pywaves.wrapper('/transactions/broadcast', data)
-            if self.pywaves.OFFLINE:
-                return req
-            else:
-                return req
+            return self.pywaves.wrapper('/transactions/broadcast', data)
 
     def invokeScript(self, dappAddress, functionName, params, payments, feeAsset=None,
                      txFee=pywaves.DEFAULT_INVOKE_SCRIPT_FEE, publicKey=None):
@@ -1301,17 +1275,17 @@ class Address(object):
                 if param['type'] == 'integer':
                     parameterBytes += b'\0' + struct.pack(">q", param['value'])
                 elif param['type'] == 'binary':
-                    parameterBytes += b'\1' + struct.pack(">I", len(param['value'])) + crypto.str2bytes(param['value'])
+                    parameterBytes += b'\1' + struct.pack(">L", len(param['value'])) + crypto.str2bytes(param['value'])
                 elif param['type'] == 'string':
                     parameterBytes += b'\2' + struct.pack(">I",
-                                                          len(crypto.str2bytes(param['value']))) + crypto.str2bytes(
-                        param['value'])
+                                                          len(crypto.str2bytes(param['value']))) + crypto.str2bytes(param['value'])
                 elif param['type'] == 'boolean':
                     if param['value'] == True:
                         parameterBytes += b'\6'
                     else:
                         parameterBytes += b'\7'
                 elif param['type'] == 'list':
+                    print(param)
                     parameterBytes += b'\x0b'
                     parameterBytes += struct.pack(">I", len(param['value']))
                     for nestedParam in param['value']:
@@ -1321,10 +1295,7 @@ class Address(object):
                             parameterBytes += b'\1' + struct.pack(">I", len(nestedParam['value'])) + crypto.str2bytes(
                                 nestedParam['value'])
                         elif nestedParam['type'] == 'string':
-                            parameterBytes += b'\2' + struct.pack(">I",
-                                                                  len(crypto.str2bytes(
-                                                                      nestedParam['value']))) + crypto.str2bytes(
-                                nestedParam['value'])
+                            parameterBytes += b'\2' + struct.pack(">I",len(crypto.str2bytes(nestedParam['value']))) + crypto.str2bytes(nestedParam['value'])
                         elif nestedParam['type'] == 'boolean':
                             if nestedParam['value'] == True:
                                 parameterBytes += b'\6'
